#ifndef FONTFORGE_IKARUS_H
#define FONTFORGE_IKARUS_H

#include "splinefont.h"

#ifdef __cplusplus
extern "C" {
#endif

extern SplineFont *SFReadIkarus(char *fontname);

#ifdef __cplusplus
}
#endif

#endif /* FONTFORGE_IKARUS_H */
