/* Copyright (C) 2025 by Maxim Iorsh <iorsh@users@sourceforge.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#pragma once

#include <map>
#include <vector>
#include <variant>

#include "l10n_text.hpp"

namespace ff::dlg {

using NumericalValue = std::variant<std::monostate, int, double>;
using ProblemRecordsOut = std::map<short /*cid*/, NumericalValue>;

struct ProblemRecord {
    short cid;
    L10nText label;
    L10nText tooltip;
    bool active;
    NumericalValue value;
    short parent_cid;
    bool disabled;
};

struct ProblemTab {
    L10nText label;
    std::vector<ProblemRecord> records;
};

/* This function updates pr_tabs in-place to preserve the state of the dialog
   between invocations.

   Return value: true, if any problem record was selected. The selected records
                 are marked as active in pr_tabs. */
bool find_problems_dialog(GWindow parent, std::vector<ProblemTab>& pr_tabs,
                          double& near);

}  // namespace ff::dlg
