#ifndef FONTFORGE_CLIPNOUI_H
#define FONTFORGE_CLIPNOUI_H

#include "uiinterface.h"

#ifdef __cplusplus
extern "C" {
#endif

extern void FF_SetClipInterface(struct clip_interface *clipi);

#ifdef __cplusplus
}
#endif

#endif /* FONTFORGE_CLIPNOUI_H */
