#ifndef FONTFORGE_SPLINEREFIGURE_H
#define FONTFORGE_SPLINEREFIGURE_H

#include "splinefont.h"

#ifdef __cplusplus
extern "C" {
#endif

extern void SplineRefigure3(Spline *spline);

#ifdef __cplusplus
}
#endif

#endif /* FONTFORGE_SPLINEREFIGURE_H */
