#ifndef FONTFORGE_CRCTAB_H
#define FONTFORGE_CRCTAB_H

#ifdef __cplusplus
extern "C" {
#endif

extern unsigned long binhex_crc(unsigned char *buffer, int size);

#ifdef __cplusplus
}
#endif

#endif /* FONTFORGE_CRCTAB_H */
