/* Copyright 2024 Maxim Iorsh <iorsh@users.sourceforge.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.

 * The name of the author may not be used to endorse or promote products
 * derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#pragma once

#include <gtkmm.h>

typedef struct gwindow* GWindow;

namespace ff::dlg {

// Modal dialog
class DialogBase : public Gtk::Dialog {
 public:
    // The parent is a legacy GDraw window.
    // TODO(iorsh): remove this constructor after the transition to GTK is
    // complete.
    DialogBase(GWindow parent_gwin);
    ~DialogBase();

    Gtk::ResponseType run();

    // Add Help context to be opened if the user presses "F1".
    void set_help_context(const std::string& file,
                          const std::string& section = "");

    // Allow window resize in horizontal direction only.
    void set_hints_horizontal_resize_only();

    // The Cancel button is added in this class as a named widget, so that it
    // can be later styled with CSS. This functions removes the Cancel button
    // from the dialog, if it is not needed.
    void remove_cancel_button();

 private:
    GWindow parent_gwindow_ = nullptr;

    std::string help_file_, help_section_;

    bool on_help_key_press(GdkEventKey* event);
};

}  // namespace ff::dlg
